function [x] = pg_rest_test(degraded, image, resD, resP, NbIt, lambda, xmin, xmax, h, varG)
%   Demonstration of all experiments presented in:
% 	
% 	[2] Emilie Chouzenoux, Anna Jezierska, Jean-Christophe Pesquet, and Hugues Talbot
% 	A Convex Approach for Image Restoration with Exact Poisson-Gaussian Likelihood,
%   submitted to SIAM Journal on Imaging Science, 2015
% 	
%
%   Copyright 2013 Anna Jezierska
% 	This function computes the Poisson-Gaussian data fidelity term
% 	as described in
% 
% 
% 	[1] Anna Jezierska, Emilie Chouzenoux, Jean-Christophe Pesquet, and Hugues Talbot
% 	A primal-dual proximal splitting approach for restoring data corrupted with Poisson-Gaussian noise,
% 	International Conference on Acoustics, Speech, and Signal Processing (ICASSP),
% 	Kyoto, 25-30 March, 2012.
% 
% 	[2] Emilie Chouzenoux, Anna Jezierska, Jean-Christophe Pesquet, and Hugues Talbot
% 	A Convex Approach for Image Restoration with Exact Poisson-Gaussian Likelihood,
%   submitted to SIAM Journal on Imaging Science, 2015
% 
% 	[3] Anna Jezierska (2013).
% 	Image Restoration in the presence of Poisson-Gaussian noise.
% 	Ph.D. Thesis. Université Paris-Est
% 
% 	This software was developed by Anna Jezierska version of 03.2014
% 	(Anna.Jezierska@ibspan.waw.pl).
% 
% 	If you use this software for research purposes, you should cite
% 	the aforementioned paper in any resulting publication.
% .
% INPUTS: 
%	degraded - name of input image (degraded), txt file saved in folder: 'Images/',degraded,'.txt'
% 	image - name of clear image, txt file saved in folder: 'Images/',image,'.txt'
%	resD - data fidelity choice, takes value between 1 and 7, where '1- Poiss\n' '2-Gauss\n' '3-GAST\n' '4-EXP\n' '5-SPoiss\n' '6-WL2\n' '7-PoissGauss\n'
%	resP - prior choice, takes value between 1 and 3, where '1-TV \n', '2-NLTV \n', '3-HTV \n' 
%	NbIt - number of iterations, sclar value
%	lambda - regularization parameter, scalar for TV and NLTV prior and vector for HTV prior
%	xmin - minimum value in resultng image, e.g. xmin = 0;
%	xmax - minimum value in resultng image, e.g. xmax = 30;
%	h - blur, e.g. h = fspecial('gaussian', 25, 1.6);
% 	varG - varaiance of Gaussian noise

addpath Tools
addpath algo
addpath NLTV_functions
addpath BM3D_functions
addpath PG_fid_functions
addpath Benvenuto_fid_functions
addpath LambertW
addpath Images

data_fidelity_str = {'Poiss' 'Gauss' 'GAST' 'EXP' 'SPoiss' 'WL2' 'PoissGauss' };
prior_str = {'TV' 'NLTV' 'HTV'};

                     
%%%%%%%%%%DATA FIDELITY%%%%%%%%%%%%%%%%%%%%
prompt = ['Data fidelity choice:\n' '1- Poiss\n' '2-Gauss\n' '3-GAST\n' '4-EXP\n' '5-SPoiss\n' '6-WL2\n' '7-PoissGauss\n' ];
%fprintf(prompt);     
%resD = input(prompt);
data_fidelity = data_fidelity_str(resD);
data_fidelity=num2str(cell2mat(data_fidelity));

%%%%%%%%%%PRIOR TERM%%%%%%%%%%%%%%%%%%%%
prior=prior_str(resP);
prior=num2str(cell2mat(prior));

%%%%%%%%%%BLUR%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%degraded=num2str(cell2mat(degraded));
%blur=num2str(cell2mat(blur));


I = importdata(strcat('Images/',image,'.txt')); % read clear image
y = importdata(strcat('Images/',degraded,'.txt')); % read degraded image
y=y(:);


%%%%%%%%%%SUMMARY%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp(['Image: ' image ]); 
disp(['Data fidelity: ' data_fidelity ]); 
disp(['Prior term: ' prior]); 
%disp(['Blur: ' blur]); 
disp(['Variance of Gaussian noise: ' num2str(varG)]); 
disp(['x+: ', num2str(xmax)]);
disp(['Lambda: ', num2str(lambda)]);
%%%%%%%%%%NOISE%%%%%%%%%%%%%%%%%%%%%%%%%%%%
alpha = 1;
m = 0;
tolerance = 3; % exact Poisson-Gaussian case
yTrunc = dataTruncation(data_fidelity, y, varG, m);
[nx,ny] = size(I);
x0 = ones(nx,ny); % initialization of x
x0 = x0(:);
switch prior
    case 'TV'
        [V1,V2] = difference_matrices_iso(nx,ny); % TV matrices
        L = [V1 ;V2];
        gamma = getGamma(data_fidelity, L, h, yTrunc, varG, m,nx,ny);
        [MAE, SNR, x] = primaldualrecPGNLTV(x0,y,yTrunc,h,NbIt,nx,ny,varG,m,alpha,lambda,xmin,xmax,I,data_fidelity,prior,L,0,gamma,tolerance);
    case 'NLTV'
        SNRinit = 10*log10(sum(I(:).^2)/sum((I(:)-y(:)).^2));
        [L, degree, x0] = NLTVoperator(y,nx,ny,h,SNRinit,xmax);
        gamma = getGamma(data_fidelity, L, h, yTrunc, varG, m,nx,ny);
        [MAE, SNR, x] = primaldualrecPGNLTV(x0,y,yTrunc,h,NbIt,nx,ny,varG,m,alpha,lambda,xmin,xmax,I,data_fidelity,prior,L,degree,gamma,tolerance);
    case 'HTV'
        [V11,V22,V12] = difference_matrices_unser(nx,ny); % Hessian matrices
        [V1,V2] = difference_matrices_iso(nx,ny); % TV matrices
        VHes =  [V11 ;sqrt(2)*V12 ;V22];
        VTV = [V1 ;V2];
        gamma = getGammaHyb(data_fidelity, VHes,VTV, h, yTrunc, varG, m,nx,ny);
        x = primaldualrecPG_Hyb(x0,y,yTrunc,h,NbIt,nx,ny,varG,m,alpha,lambda(2),lambda(1), xmin,xmax,I,data_fidelity,prior, VHes, VTV, 0,gamma,tolerance);
end

end

